"""
:mod:`zsl.db.helpers.pagination`
--------------------------------
"""
from __future__ import unicode_literals
from builtins import object
DEFAULT_PAGE_RECORD_COUNT = 25


class Pagination(object):
    def __init__(self, pagination):
        if 'offset' in pagination:
            self._offset = int(pagination['offset'])
        else:
            self._offset = 0

        if 'page_record_count' in pagination:
            self._page_record_count = int(pagination['page_record_count'])
        else:
            self._page_record_count = DEFAULT_PAGE_RECORD_COUNT

        self._record_count = 0

    def set_record_count(self, record_count):
        self._record_count = record_count

    def get_record_count(self):
        return self._record_count

    def get_page_record_count(self):
        return self._page_record_count

    def get_offset(self):
        if self._offset >= self.get_record_count():
            self._offset = max(self.get_record_count() - 1, 0)

        return self._offset

    def apply_pagination(self, q):
        return q.limit(self.get_page_record_count()).offset(self.get_offset())

    def paginate(self, q):
        self.set_record_count(q.count())
        return self.apply_pagination(q).all()



"""
:mod:`zsl.interface.webservice.utils.error_handler`
---------------------------------------------------

.. moduleauthor:: Martin Babka
"""
from __future__ import unicode_literals
from builtins import str
from builtins import object

import traceback
import logging
from functools import wraps
from abc import abstractmethod

from flask import request

_error_handlers = []


class ErrorHandler(object):
    @abstractmethod
    def can_handle(self, e):
        pass

    @abstractmethod
    def handle(self, e):
        pass


def register(e):
    # type: (ErrorHandler)->None
    _error_handlers.append(e)


def error_handler(f):
    """
    Default error handler.
     - On server side error shows a message 'An error occurred!' and returns 500 status code.
     - Also serves well in the case when the resource/task/method is not found - returns 404 status code.
    """

    @wraps(f)
    def error_handling_function(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except ImportError as ie:
            logging.error(str(ie) + "\n" + traceback.format_exc())
            return str(ie), 404
        except Exception as ex:
            for eh in _error_handlers:
                if eh.can_handle(ex):
                    return eh.handle(ex)

            logging.error(str(ex) + "\n" + traceback.format_exc())
            logging.error("Request:\n{0}\n{1}\n".format(request.headers, request.data))

            return "An error occurred!", 500

    return error_handling_function



"""
:mod:`zsl.utils.reflection_helper`
----------------------------------

Helper module for OOP operations.

.. moduleauthor:: Martin Babka
"""
from __future__ import unicode_literals


def extend(instance, new_class):
    """

    :param instance:
    :param new_class:
    :return:
    """
    instance.__class__ = type(
        '%s_extended_with_%s' % (instance.__class__.__name__, new_class.__name__),
        (new_class, instance.__class__,),
        {}
    )


def add_mixin(base_class, mixin_class):
    base_class.__bases__ = (mixin_class,) + base_class.__bases__
    return base_class




"""
:mod:`zsl.interface.webservice.performers.task`
-----------------------------------------------
"""
from __future__ import unicode_literals
from builtins import str
from flask import request
from flask.helpers import make_response

from zsl import inject, Zsl, Injected
from zsl.interface.webservice.utils.request_data import extract_data
from zsl.task.job_context import JobContext, WebJobContext
from zsl.interface.webservice.utils.error_handler import error_handler
from zsl.interface.webservice.utils.response_headers import append_headers
from zsl.router.task import TaskRouter


@inject(app=Zsl)
def create_web_task(app):
    @app.route("/task/<path:path>", methods=["POST", "GET", "OPTIONS"])
    @append_headers
    @error_handler
    @inject(task_router=TaskRouter)
    def perform_web_task(path, task_router=Injected):
        app.logger.debug("Performing task %s.", path)
        if request.method == 'OPTIONS':
            response = app.make_default_options_response()
        else:
            (task, task_callable) = task_router.route(path)
            data = extract_data(request)
            app.logger.debug("Data found '%s'.", str(data))
            jc = WebJobContext(path, data, task, task_callable, request)
            JobContext.set_current_context(jc)
            response = make_response(task_callable(jc.task_data))
            jc.notify_responders(response)

        return response



"""
:mod:`zsl.task.job_context`
---------------------------

.. moduleauthor:: Martin Babka <babka@atteq.com>
"""
from __future__ import unicode_literals
from builtins import object
from typing import Callable
from abc import abstractmethod

from zsl.task.task_data import TaskData


class Job(object):
    def __init__(self, data):
        # type: (dict) -> None

        self.data = data

    @property
    def path(self):
        """Job's path.

        :getter: Returns job's path
        :type: str
        """
        return self.data['path']

    @property
    def payload(self):
        """Data part of job.

        :getter: Returns job's payload
        :type: dict
        """
        return self.data['data']

    @property
    def is_valid(self):
        """Validity of job's data.

        :getter: Returns if job's data are valid
        :type: bool
        """
        return self.data and 'path' in self.data and 'data' in self.data


class JobContext(object):
    """Job Context"""

    _current_context = None

    def __init__(self, job, task, task_callable):
        # type: (Job, object, Callable) -> None
        """
        Constructor
        """
        self.job = job
        self.task = task
        self.task_callable = task_callable
        self.task_data = TaskData(self.job.payload)

        JobContext.set_current_context(None)

    @classmethod
    def get_current_context(cls):
        return cls._current_context

    @classmethod
    def set_current_context(cls, context):
        cls._current_context = context


class Responder(object):

    @abstractmethod
    def respond(self, r):
        pass


def web_task_redirect(location):
    return {'headers': {'Location': location}, 'status_code': 301}


class WebJobContext(JobContext):
    def __init__(self, path, data, task, task_callable, request):
        """Constructor"""

        self.job = Job({'data': data, 'path': path})
        self.task = task
        self.task_callable = task_callable
        self.task_data = TaskData(data)
        self._request = request
        self._responders = []

    def get_web_request(self):
        return self._request

    def add_responder(self, r):
        self._responders.append(r)

    def notify_responders(self, response):
        for r in self._responders:
            r.respond(response)



"""
:mod:`zsl.interface.webservice.utils.response_headers`
------------------------------------------------------

.. moduleauthor:: Martin Babka <babka@atteq.com>
"""
from __future__ import unicode_literals
from flask import Response
from flask.helpers import make_response
from functools import wraps

from zsl import inject, Zsl, Config, Injected


@inject(config=Config)
def append_crossdomain(response, config=Injected):
    # The CORS has already been setup.
    if 'Access-Control-Allow-Origin' in response.headers:
        return

    if config.get('ALLOW_ORIGIN'):
        response.headers['Access-Control-Allow-Origin'] = \
            config.get('ALLOW_ORIGIN')

    response.headers['Access-Control-Expose-Headers'] = \
        'location, x-total-count, link'
    response.headers['Access-Control-Allow-Methods'] = \
        'POST, GET, OPTIONS, PUT, PATCH, DELETE'
    response.headers['Access-Control-Allow-Headers'] = \
        'accept, origin, content-type, authorization'


@inject(app=Zsl)
def append_asl(response, app=Injected):
    response.headers['ZSL-Flask-Layer'] = app.get_version()


def append_cache(response):
    response.headers['Cache-Control'] = 'no-cache'


def append_all(response):
    append_asl(response)
    append_cache(response)
    append_crossdomain(response)


def append_headers(f):

    @wraps(f)
    def _response_decorator(*args, **kwargs):
        r = f(*args, **kwargs)
        response = r if isinstance(r, Response) else make_response(r)
        append_all(response)
        return response

    return _response_decorator



"""
:mod:`zsl.interface.webservice.performers.method`
-------------------------------------------------

.. moduleauthor:: Martin Babka
"""
from __future__ import unicode_literals
from zsl.router.method import get_method_packages
import importlib


def call_exposers_in_method_packages():
    for package in get_method_packages():
        module = importlib.import_module(package)
        if hasattr(module, '__exposer__'):
            getattr(module, '__exposer__')()



"""
:mod:`zsl.testing.db`
---------------------
This module allows for database unit testing. For how to use the database 
testing in practice, a sample, refer to :ref:`unit-testing-db`.
 
The module works in the following way (methods setUp, tearDown):
1. Each test runs in a single transaction.
2. This transaction is always called a rollback.

All the tests are run in a single parent transaction (setUpClass, 
tearDownClass):
1. In general initialization phase the session/transaction is created
and it is kept during all the testing. Also the database schema is created.
2. After this the transaction is called rollback.

This means that the tests may be conducted in the in the memory database 
or a persistent one which is kept clean.

The module provides class :class:`.TestSessionFactory` - it always returns
the same session. Also one should add :class:`.DbTestModule` to the test 
container when creating Zsl instance, see :ref:`unit-testing-zsl-instance`.
"""
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *

from injector import Module, provides, singleton
from sqlalchemy.engine import Engine
from sqlalchemy.orm.session import Session

from zsl import inject
from zsl.db.model.sql_alchemy import metadata
from zsl.service.service import SessionFactory


class TestSessionFactory(SessionFactory):
    """Factory always returning the single test transaction."""
    _test_session = None

    def __init__(self):
        super(TestSessionFactory, self).__init__()

    def create_session(self):
        # type: () -> Session
        if TestSessionFactory._test_session is None:
            TestSessionFactory._test_session = self._session_holder()

        assert TestSessionFactory._test_session is not None
        return TestSessionFactory._test_session

    def close_session(self):
        pass


class DbTestModule(Module):
    """Module fixing the :class:`zsl.service.service.SessionFactory`
    to our :class:`.TestSessionFactory`."""

    @provides(SessionFactory, scope=singleton)
    def get_session_factory(self):
        # type: ()->SessionFactory
        return TestSessionFactory()

    @provides(TestSessionFactory, scope=singleton)
    @inject(session_factory=SessionFactory)
    def get_test_session_factory(self, session_factory):
        # type: (SessionFactory)->SessionFactory
        return session_factory


class DbTestCase(object):
    """:class:`.DbTestCase` is a mixin to be used when testing with a database."""

    _session = None

    @classmethod
    @inject(session_factory=TestSessionFactory, engine=Engine)
    def setUpClass(cls, engine, session_factory):
        # type: (Engine, SessionFactory)->None
        super(DbTestCase, cls).setUpClass()
        DbTestCase._session = session_factory.create_session()
        metadata.bind = engine
        metadata.create_all(engine)

    @inject(session_factory=TestSessionFactory)
    def setUp(self, session_factory):
        # type: (SessionFactory)->None
        session = session_factory.create_session()
        session.begin(subtransactions=True)

    @inject(session_factory=TestSessionFactory)
    def tearDown(self, session_factory):
        # type: (SessionFactory)->None
        # This will return the same transaction/session
        # as the one used in setUp.
        sess = session_factory.create_session()
        sess.rollback()
        sess.close()

    @classmethod
    def tearDownClass(cls):
        session = DbTestCase._session
        session.rollback()
        session.close()




"""
:mod:`zsl.application.modules.alchemy_module`
---------------------------------------------
"""
from __future__ import unicode_literals
from builtins import object

import logging

from sqlalchemy import create_engine
from sqlalchemy.orm.session import sessionmaker, Session
from sqlalchemy.engine.base import Engine

from injector import Module, provides, singleton, inject

from zsl import Config


class SessionHolder(object):
    def __init__(self, sess_cls):
        self._sess_cls = sess_cls

    def __call__(self):
        # type: () -> Session
        return self._sess_cls()


class AlchemyModule(Module):
    """Adds SQLAlchemy to current configuration."""
    @provides(Engine, scope=singleton)
    @inject(config=Config)
    def provide_engine(self, config):
        # type: (Config) -> Engine

        engine = create_engine(config['DATABASE_URI'], **config['DATABASE_ENGINE_PROPS'])
        logging.debug("Created DB configuration to {0}.".format(config['DATABASE_URI']))

        return engine

    @provides(SessionHolder, scope=singleton)
    @inject(engine=Engine)
    def provide_session_holder(self, engine):
        # type: (Engine) -> SessionHolder

        session = SessionHolder(sessionmaker(engine, autocommit=False, expire_on_commit=False))
        logging.debug("Created ORM session")

        return session



"""
:mod:`zsl.task.task_data`
-------------------------

.. moduleauthor:: Martin Babka <babka@atteq.com>
"""
from __future__ import unicode_literals
from builtins import object

from zsl import inject, Injected, Zsl


class TaskData(object):

    @inject(app=Zsl)
    def __init__(self, data, app=Injected):
        self._app = app
        self._data = data
        self._is_skipping_json = False

    def get_data(self):
        return self._data

    # TODO should this be here? If some component needs the app instance, it
    # should ask for it
    def get_service_application(self):
        return self._app

    def transform_data(self, f):
        self._data = f(self._data) if self._data is not None else {}

    def is_skipping_json(self):
        return self._is_skipping_json

    def set_skipping_json(self, value):
        self._is_skipping_json = value



"""
:mod:`zsl.db.model.app_model_json_encoder`
------------------------------------------

.. moduleauthor:: Martin Babka
"""
from __future__ import unicode_literals
from json.encoder import JSONEncoder
from zsl.db.model import AppModel


class AppModelJSONEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, AppModel):
            d = dict(o.__dict__)
            for k in o._not_serialized_attributes:
                if k in d:
                    d.pop(k)
            return d
        else:
            return JSONEncoder.default(self, o)



import re
import ast
import sys
from setuptools import setup, find_packages

EXCLUDE_FROM_PACKAGES = ['tests', 'tests.*']

_version_re = re.compile(r'__version__\s+=\s+(.*)')

with open('zsl/__init__.py', 'rb') as f:
    version = str(ast.literal_eval(_version_re.search(
        f.read().decode('utf-8')).group(1)))

_is_py3 = sys.version_info > (3, 0)

setup(name='zsl',
      version=version,
      description='zsl application framework for web based services',
      author='Atteq s.r.o.',
      author_email='open.source@atteq.com',
      url='https://github.com/AtteqCom/zsl',
      license='MIT',
      install_requires=[
          'future',
          'Flask-Injector>=0.8' if _is_py3 else 'Flask-Injector==0.8.0',
          'requests>=2.7',
          'SQLAlchemy>=1',
          'typing>=3.5',
          'Werkzeug>=0.12',
      ],
      extras_require={
          'cli': ['bpython'],
          'redis': ['redis>=2.10'],
          'celery': ['zsl_client'],
          'gearman': ['zsl_client', 'gearman'],
          'alembic': ['alembic'],
          'documentation': ['sphinx', 'recommonmark', 'sphinx_rtd_theme',
                            'alembic']
      },
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3.6',
          'Operating System :: OS Independent',
          'Topic :: Software Development :: Libraries :: Application Frameworks'
      ],
      packages=find_packages(exclude=EXCLUDE_FROM_PACKAGES))



"""
:mod:`zsl.task.task_decorator`
------------------------------

.. moduleauthor:: Martin Babka <babka@atteq.com>, 
                  Peter Morihladko <morihladko@atteq.com>, 
                  Jan Janco <janco@atteq.com>,
                  Lubos Pis <pis@atteq.com>
"""
from __future__ import unicode_literals

from builtins import bool
from builtins import str

import logging
import json

from datetime import timedelta
from typing import Callable

from flask import request
from flask.config import Config

from zsl.constants import MimeType
from zsl.task.task_data import TaskData
from zsl.db.model import AppModelJSONEncoder
from zsl.task.job_context import JobContext, WebJobContext, Responder
import traceback
from functools import wraps
from os.path import os
from zsl.utils.file_helper import makedirs
from zsl.utils.security_helper import verify_security_data

from zsl import inject, Zsl, Injected


def log_output(f):
    """
    Logs the output value.
    """

    @wraps(f)
    def wrapper_fn(*args, **kwargs):
        res = f(*args, **kwargs)
        logging.debug("Logging result %s.", res)
        return res

    return wrapper_fn


def save_to_file(destination_filename, append=False):
    """
    Save the output value to file.
    """

    def decorator_fn(f):
        @wraps(f)
        def wrapper_fn(*args, **kwargs):
            res = f(*args, **kwargs)

            makedirs(os.path.dirname(destination_filename))
            mode = "a" if append else "w"
            with open(destination_filename, mode) as text_file:
                text_file.write(res)

            return res

        return wrapper_fn

    return decorator_fn


def json_input(f):
    """
    Expects task input data in json format and parse this data.
    """

    @wraps(f)
    def json_input_decorator(*args, **kwargs):
        # If the data is already transformed, we do not transform it any
        # further.
        task_data = _get_data_from_args(args)

        if task_data is None:
            logging.error("Task data is empty during JSON decoding.")

        if task_data.get_data():
            try:
                is_transformed = request.json
                should_transform = not task_data.is_skipping_json() if \
                    task_data is not None else False

                # We transform the data only in the case of plain POST requests.
                if not is_transformed and should_transform:
                    task_data.transform_data(json.loads)
            except (ValueError, RuntimeError):
                logging.error(
                    "Exception while processing JSON input decorator.")
                task_data.transform_data(json.loads)
        else:
            task_data.transform_data(lambda _: {})

        return f(*args, **kwargs)

    return json_input_decorator


def json_output(f):
    """
    Format response to json and in case of web-request set response content type
    to 'application/json'.
    """

    @wraps(f)
    def json_output_decorator(*args, **kwargs):
        rv = f(*args, **kwargs)

        skip_encode = False
        for d in args:
            if isinstance(d, TaskData):
                skip_encode = d.is_skipping_json()

        if not skip_encode:
            rv = json.dumps(rv, cls=AppModelJSONEncoder)
            if isinstance(JobContext.get_current_context(), WebJobContext):
                JobContext.get_current_context().add_responder(
                    MimeSetterWebTaskResponder(MimeType.APPLICATION_JSON.value))

        return rv

    return json_output_decorator


def jsonp_wrap(callback_key='callback'):
    """
    Format response to jsonp and add a callback to JSON data - a jsonp request
    """

    def decorator_fn(f):

        @wraps(f)
        def jsonp_output_decorator(*args, **kwargs):
            task_data = _get_data_from_args(args)
            data = task_data.get_data()

            if callback_key not in data:
                raise KeyError(
                    'Missing required parameter "{0}" for task.'.format(
                        callback_key))

            callback = data[callback_key]
            jsonp = f(*args, **kwargs)
            if isinstance(JobContext.get_current_context(), WebJobContext):
                JobContext.get_current_context().add_responder(
                    MimeSetterWebTaskResponder('application/javascript'))
            jsonp = "{callback}({data})".format(callback=callback, data=jsonp)

            return jsonp

        return jsonp_output_decorator

    return decorator_fn


def jsend_output(fail_exception_classes=None):
    """
    Format task result to json output in jsend specification format. See: 
    http://labs.omniti.com/labs/jsend. Task return value must be dict or None.

    @param fail_exception_classes: exceptions which will produce 'fail' response
    status.
    """

    fail_exception_classes = fail_exception_classes if fail_exception_classes \
        else ()

    def decorator_fn(f):

        @wraps(f)
        @json_output
        def jsend_output_decorator(*args, **kwargs):
            try:
                rv = f(*args, **kwargs)
            except fail_exception_classes as e:
                return {'status': 'fail', 'data': {'message': str(e)}}
            except Exception as e:
                logging.error(str(e) + "\n" + traceback.format_exc())
                return {'status': 'error', 'message': 'Server error.'}

            if not isinstance(rv, dict) and rv is not None:
                msg = 'jsend_output decorator error: task must return dict ' \
                      'or None.\nTask return value: {0}.'
                logging.error(msg.format(rv))
                return {'status': 'error', 'message': 'Server error.'}

            return {'status': 'success', 'data': rv}

        return jsend_output_decorator

    return decorator_fn


def web_error_and_result(f):
    """
    Same as error_and_result decorator, except:
    If no exception was raised during task execution, ONLY IN CASE OF WEB 
    REQUEST formats task result into json dictionary {'data': task return value}
    """

    @wraps(f)
    def web_error_and_result_decorator(*args, **kwargs):
        return error_and_result_decorator_inner_fn(f, True, *args, **kwargs)

    return web_error_and_result_decorator


def error_and_result(f):
    """
    Format task result into json dictionary `{'data': task return value}` if no 
    exception was raised during the task execution. If there was raised an 
    exception during task execution, formats task result into dictionary
    `{'error': exception message with traceback}`.
    """

    @wraps(f)
    def error_and_result_decorator(*args, **kwargs):
        return error_and_result_decorator_inner_fn(f, False, *args, **kwargs)

    return error_and_result_decorator


def error_and_result_decorator_inner_fn(f, web_only, *args, **kwargs):
    try:
        ret_val = f(*args, **kwargs)
        if web_only and not isinstance(JobContext.get_current_context(),
                                       WebJobContext):
            rv = ret_val
        else:
            rv = {'data': ret_val}
    except:
        exc = traceback.format_exc()
        logging.error(exc)
        rv = {'error': "{0}".format(exc)}

    return json.dumps(rv)


def required_data(*data):
    """
    Task decorator which checks if the given variables (indices) are stored 
    inside the task data.
    """

    def decorator_fn(f):

        @wraps(f)
        def required_data_decorator(*args, **kwargs):
            task_data = _get_data_from_args(args).get_data()
            for i in data:
                if i not in task_data:
                    raise KeyError(
                        'Missing required parameter "{0}" for task.'.format(i))

            return f(*args, **kwargs)

        return required_data_decorator

    return decorator_fn


def append_get_parameters(accept_only_web=True):
    # type: (bool) -> Callable
    """
    Task decorator which appends the GET data to the task data.

    :param accept_only_web: Parameter which limits using this task only 
                            with web requests.
    """

    def wrapper(f):

        @wraps(f)
        def append_get_parameters_wrapper_fn(*args, **kwargs):
            jc = JobContext.get_current_context()

            if isinstance(jc, WebJobContext):
                # Update the data with GET parameters
                web_request = jc.get_web_request()
                task_data = _get_data_from_args(args)
                data = task_data.get_data()
                data.update(web_request.args.to_dict(flat=True))
            elif accept_only_web:
                # Raise exception on non web usage if necessary
                raise Exception("append_get_parameters decorator may be used "
                                "with GET requests only.")

            return f(*args, **kwargs)

        return append_get_parameters_wrapper_fn

    return wrapper


def web_task(f):
    """
    Checks if the task is called through the web interface.
    Task return value should be in format {'data': ...}.
    """

    @wraps(f)
    def web_task_decorator(*args, **kwargs):
        jc = JobContext.get_current_context()
        if not isinstance(jc, WebJobContext):
            raise Exception(
                "The WebTask is not called through the web interface.")
        data = f(*args, **kwargs)
        jc.add_responder(WebTaskResponder(data))
        return data['data'] if 'data' in data else ""

    return web_task_decorator


def secured_task(f):
    """
    Secured task decorator.
    """

    @wraps(f)
    def secured_task_decorator(*args, **kwargs):
        task_data = _get_data_from_args(args)
        assert isinstance(task_data, TaskData)
        if not verify_security_data(task_data.get_data()['security']):
            raise SecurityException(
                task_data.get_data()['security']['hashed_token'])

        task_data.transform_data(lambda x: x['data'])
        return f(*args, **kwargs)

    return secured_task_decorator


def xml_output(f):
    """
    Set content-type for response to WEB-REQUEST to 'text/xml'
    """

    @wraps(f)
    def xml_output_inner_fn(*args, **kwargs):
        ret_val = f(*args, **kwargs)

        if isinstance(JobContext.get_current_context(), WebJobContext):
            JobContext.get_current_context().add_responder(
                MimeSetterWebTaskResponder('text/xml'))
        return ret_val

    return xml_output_inner_fn


def file_upload(f):
    """
    Return list of `werkzeug.datastructures.FileStorage` objects - files to be
    uploaded
    """

    @wraps(f)
    def file_upload_decorator(*args, **kwargs):
        # If the data is already transformed, we do not transform it any
        # further.
        task_data = _get_data_from_args(args)

        if task_data is None:
            logging.error("Task data is empty during FilesUploadDecorator.")

        task_data.transform_data(lambda _: request.files.getlist('file'))
        return f(*args, **kwargs)

    return file_upload_decorator


def _get_data_from_args(args):
    task_data = None
    for d in args:
        if isinstance(d, TaskData):
            task_data = d

    return task_data


class SecurityException(Exception):
    def __init__(self, hashed_token):
        Exception.__init__(self,
                           "Invalid hashed token '{0}'.".format(hashed_token))
        self._hashed_token = hashed_token

    def get_hashed_token(self):
        return self._hashed_token


class WebTaskResponder(Responder):
    def __init__(self, data):
        self.data = data

    def respond(self, response):
        for k in self.data:
            if k == 'headers':
                for header_name in self.data[k]:
                    response.headers[header_name] = self.data[k][header_name]
            else:
                setattr(response, k, self.data[k])


class MimeSetterWebTaskResponder(Responder):
    def __init__(self, mime):
        self._mime = mime

    def respond(self, r):
        r.content_type = self._mime


class CrossdomainWebTaskResponder(Responder):
    """
    source: http://flask.pocoo.org/snippets/56/
    """

    @inject(app=Zsl, config=Config)
    def __init__(self, origin=None, methods=None, headers=None, max_age=21600,
                 app=Injected, config=Injected):
        self._app = app

        if methods is not None:
            methods = ', '.join(sorted(x.upper() for x in methods))
        self.methods = methods

        if headers is not None and not isinstance(headers, (str, bytes)):
            headers = ', '.join(x.upper() for x in headers)
        self.headers = headers

        if origin is None:
            origin = config.get('CORS', {}).get('origin', '')

        if not isinstance(origin, (str, bytes)):
            origin = ', '.join(origin)
        self.origin = origin

        if isinstance(max_age, timedelta):
            max_age = max_age.total_seconds()
        self.max_age = max_age

    def get_methods(self):
        if self.methods is not None:
            return self.methods

        options_resp = self._app.make_default_options_response()
        return options_resp.headers['allow']

    def respond(self, resp):
        h = resp.headers

        h['Access-Control-Allow-Origin'] = self.origin
        h['Access-Control-Allow-Methods'] = self.get_methods()
        h['Access-Control-Max-Age'] = str(self.max_age)
        if self.headers is not None:
            h['Access-Control-Allow-Headers'] = self.headers
        return resp


def crossdomain(origin=None, methods=None, headers=None, max_age=21600):
    def decorator(f):
        responder = CrossdomainWebTaskResponder(origin, methods, headers,
                                                max_age)

        @wraps(f)
        def crossdomain_inner_fn(*args, **kwargs):
            rv = f(*args, **kwargs)
            if isinstance(JobContext.get_current_context(), WebJobContext):
                JobContext.get_current_context().add_responder(responder)
            return rv

        return crossdomain_inner_fn

    return decorator



"""
:mod:`zsl.interface.task`
-------------------------

.. moduleauthor:: Peter Morihladko
"""
from __future__ import unicode_literals
from future.builtins import str

import json

from zsl import inject, Injected
from zsl.router.task import TaskRouter
from zsl.task.job_context import JobContext, Job


@inject(task_router=TaskRouter)
def exec_task(task_path, data, task_router=Injected):
    """Execute task.

    :param task_path: task path
    :type task_path: str
    :param data: task's data
    :type data: Any
    :param task_router: task router, injected
    :type task_router: TaskRouter
    :return:
    """
    if not data:
        data = {'data': None, 'path': task_path}

    elif not isinstance(data, (str, bytes)):
        data = json.dumps(data)

    # Open the data from file, if necessary.
    elif data is not None and data.startswith("file://"):
        with open(data[len("file://"):]) as f:
            data = f.read()

    # Prepare the task.
    job = Job(data)
    (task_, task_callable) = task_router.route(task_path)
    jc = JobContext(job, task_, task_callable)
    JobContext.set_current_context(jc)

    # Run the task.
    return jc.task_callable(jc.task_data)
